import javax.media.j3d.*;
import javax.vecmath.*;
import java.util.*;

/**
 * Title: ModellViewer
 * Description: The ModellViewer Class is used to visualize the 3D
 * representation of the Hive. It is possible to add Honeycombs and Bees.
 * @author Dimitri Missoh & Daniel Schlosser
 * @version 0.1
 */
public class ModellViewer
{
    Canvas3D cv;
    MyUniverse universe;
    double xAngle,yAngle,zAngle;
    int xTrans, yTrans ,zTrans;
    Vector3d aVector;
    Transform3D pufferTransform3D, mulTransform3D;
    BranchGroup  alltogether, HoneycombAddPoint, BeesAddPoint;
    LinkedList temporaryRemoved;
    MarkViewDialog markviewd;
    private PickDragBehavior PDB;
    private Honeycomb firstCell;
    private Koord3D crystalisationPoint=new Koord3D();

    public void initDialogConection(MarkViewDialog mv)
    {
	markviewd=mv;
    }
    public int getNumberOfHoneycombs()
    {
	return this.HoneycombAddPoint.numChildren();
    }
    public Koord3D getSize()
    {
	int n=this.HoneycombAddPoint.numChildren();
	int xmin=100,ymin=100,zmin=100;
	int xmax=-10,ymax=-10,zmax=-10;
	int x,y,z;
	Koord3D coord;
	for(int i =0;i<n;i++)
	{
	    coord=((Honeycomb)HoneycombAddPoint.getChild(i)).getpos();
	    x=coord.getX();
	    y=coord.getY();
	    z=coord.getZ();
	    if(xmin>x) xmin=x;
	    if(ymin>y) ymin=y;
	    if(zmin>z) zmin=z;
	    if(xmax<x) xmax=x;
	    if(ymax<y) ymax=y;
	    if(zmax<z) zmax=z;
	}
	return new Koord3D(xmax-xmin,ymax-ymin,zmax-zmin);
    }
    public double getDichte(long Bloecke)
    {
	Koord3D s=getSize();
	double size=s.getX()*s.getY()*s.getZ();
	if (size==0) return 0;
	return (Bloecke/size);
    }
    public ModellViewer(Canvas3D cv,TheHive theHive)
    {
	this.cv=cv;
        aVector=new Vector3d();
        xAngle=0;
        yAngle=0;
        zAngle=0;
        xTrans=0;
        yTrans=0;
        zTrans=0;
        pufferTransform3D= new Transform3D();
        mulTransform3D=new Transform3D();
        temporaryRemoved=new LinkedList();
        universe=new MyUniverse(cv,theHive);
        HoneycombAddPoint=new BranchGroup();
        HoneycombAddPoint.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
        HoneycombAddPoint.setCapability(BranchGroup.ALLOW_CHILDREN_READ);
        HoneycombAddPoint.setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
        BeesAddPoint=new BranchGroup();
        BeesAddPoint.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
        BeesAddPoint.setCapability(BranchGroup.ALLOW_CHILDREN_READ);
        BeesAddPoint.setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
        alltogether=new BranchGroup();
        alltogether.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
	alltogether.setCapability(BranchGroup.ALLOW_CHILDREN_READ);
	alltogether.setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
	alltogether.addChild(HoneycombAddPoint);
	alltogether.addChild(BeesAddPoint);
	PDB=new PickDragBehavior(cv,HoneycombAddPoint,this);
	 BoundingSphere bounds=new BoundingSphere(new Point3d(0.0,0.0,0.0),Double.POSITIVE_INFINITY);
	PDB.setSchedulingBounds(bounds);
	BoundingLeaf boundingLeaf=new BoundingLeaf(bounds);
	alltogether.addChild(boundingLeaf);
	alltogether.addChild(PDB);
        universe.addBranchGraph(alltogether);
    }

    public void setPositionOfFirstHoneycomb(Koord3D coordinate)
    {
        this.removeCell(crystalisationPoint);
	crystalisationPoint=new Koord3D(coordinate.getX(),coordinate.getY(),coordinate.getZ());
	this.addCell(crystalisationPoint);
    }

    public void reportMarkingOperationToDialogs(Koord3D coordinate)
    {
	markviewd.honeycombGotMarked(coordinate);
    }

    //fgt Wabe hinzu
    public void addCell(Koord3D cell)
    {
	if (cell==null) return;
	Honeycomb addThis=new Honeycomb(cell.getX(),cell.getY(),cell.getZ(),this);
	HoneycombAddPoint.addChild(addThis);
    }
    public void addCell(Koord3D[] waben)
    {
        if (waben==null) return;
        Honeycomb addThis;
        for(int i=0;i<waben.length;i++)
        {
	    if(waben[i]==null) continue;
            addThis=new Honeycomb(waben[i].getX(),waben[i].getY(),waben[i].getZ(),this);
            HoneycombAddPoint.addChild(addThis);
        }
    }
    // entfernt Wabe
    public void removeCell(Koord3D[] waben)
    {
        if(waben==null) return;
        for(int i=0;i<waben.length;i++)
        {
	    if(waben[i]==null) continue;
	    removeCell(waben[i]);
        }
    }

    public void removeCell(Koord3D cell)
    {
	if(cell==null) return;
	for(int i=HoneycombAddPoint.numChildren()-1;i>=0;i--)
	{
	    if(  cell.equals( ((Honeycomb)HoneycombAddPoint.getChild(i)).getpos() )  )
		 HoneycombAddPoint.removeChild(i);
	}
    }

    public void addBee(Koord3D pos)
    {
	if(pos==null) return;
        Bee3DModell aBee=new Bee3DModell(pos.getX() ,pos.getY(),pos.getZ());
        BeesAddPoint.addChild(aBee);
    }

    public void addBee(Koord3D[] pos)
    {
        if(pos==null)return;
        for(int i=0;i<pos.length;i++)
            if(pos[i]!=null)
		addBee(pos[i]);
    }
    //verschiebt die Ameise
    public void moveBees(Koord3D[] pos)
    {
         if(pos==null) return;
         int grenze=Math.min(pos.length,BeesAddPoint.numChildren());
         for(int i=0;i<grenze;i++)
         {
             ( (Bee3DModell) BeesAddPoint.getChild(i)).move(pos[i]);
         }
    }
    //rotiert um X-Achse
    public void rotateAbsolutX(double a)
    {
        universe.rotateXAbsolut(2.0*Math.PI-a);
    }
    public void rotateAbsolutY(double a)
    {
        universe.rotateYAbsolut(2.0*Math.PI-a);
    }
    public void rotateAbsolutZ(double a)
    {
        universe.rotateZAbsolut(2.0*Math.PI-a);
    }
    public void rotateRelativToView(int dx, int dy)
    {
	universe.rotateRelativToView(dx,dy);
    }
    public void translateAbsolutX(int t)
    {
        this.xTrans=t;
        universe.translateCenterAbsolut(xTrans,yTrans,zTrans);
    }
    public void translateAbsolutY(int t)
    {
        this.yTrans=t;
        universe.translateCenterAbsolut(xTrans,yTrans,zTrans);
    }
    public void translateAbsolutZ(int t)
    {
        this.zTrans=t;
        universe.translateCenterAbsolut(xTrans,yTrans,zTrans);
    }
    public void translateToCoordinates(Koord3D coordinate)
    {
	this.demarkHoneycomb(new Koord3D(xTrans,yTrans,zTrans));
	this.xTrans=coordinate.getX();
	this.yTrans=coordinate.getY();
	this.zTrans=coordinate.getZ();
	this.markHoneycomb(coordinate,java.awt.Color.white);
	universe.translateCenterAbsolut(xTrans,yTrans,zTrans);
    }
    public void zoomAbsolut(double t)
    {
        double camerawidth=100/t +1;
        universe.zoomAbsolut(camerawidth);
    }
    public void zoomRelativToView(int dy)
    {
	universe.zoomRelativ(dy);
    }

    public void markHoneycomb(Koord3D pos,java.awt.Color col)
    {
        if(HoneycombAddPoint.numChildren()==0)return;
        for(int i=HoneycombAddPoint.numChildren()-1;i>=0;i--)
                if(
                    pos.equals(
                                ((Honeycomb) HoneycombAddPoint.getChild(i)).getpos()
                              )
                  )
                    ((Honeycomb) HoneycombAddPoint.getChild(i)).setMarked(true,col);
    }

    public void demarkHoneycomb(Koord3D pos)
    {
        if(HoneycombAddPoint.numChildren()==0)return;
        for(int i=HoneycombAddPoint.numChildren()-1;i>=0;i--)
                if(
                    pos.equals(
                                ((Honeycomb) HoneycombAddPoint.getChild(i)).getpos()
                              )
                  )
                    ((Honeycomb) HoneycombAddPoint.getChild(i)).setMarked(false,java.awt.Color.gray);
    }

    public void changeHoneycombColor(Koord3D pos,java.awt.Color col)
    {
        if(HoneycombAddPoint.numChildren()==0)return;
        for(int i=HoneycombAddPoint.numChildren()-1;i>=0;i--)
                if(
                    pos.equals(
                                ((Honeycomb) HoneycombAddPoint.getChild(i)).getpos()
                              )
                  )
                    ((Honeycomb) HoneycombAddPoint.getChild(i)).changeColor(col);
    };
    //lscht alle Waben bis auf die allererste
    public void clearall()
    {
        while(HoneycombAddPoint.numChildren()!=0)
            HoneycombAddPoint.removeChild(0);
	HoneycombAddPoint.addChild(new Honeycomb(crystalisationPoint.getX(),crystalisationPoint.getY(),crystalisationPoint.getZ(),this));
        while(BeesAddPoint.numChildren()!=0)
            BeesAddPoint.removeChild(0);
    }
    //Verndert Erscheinungsbild von selektierten Waben
    public void selectedToLines()
    {
        if(HoneycombAddPoint.numChildren()==0)return;
        for(int i=HoneycombAddPoint.numChildren()-1;i>=0;i--)
                if( ((Honeycomb) HoneycombAddPoint.getChild(i)).isSelected())
		    ((Honeycomb) HoneycombAddPoint.getChild(i)).toLines(true);
    }
    //Lscht selektierte Waben temporr;
    public void selectedRemove()
    {
        if(HoneycombAddPoint.numChildren()==0)return;
        for(int i=HoneycombAddPoint.numChildren()-1;i>=0;i--)
            if( ((Honeycomb) HoneycombAddPoint.getChild(i)).isSelected())
                temporaryRemoved.add( ((Honeycomb) HoneycombAddPoint.getChild(i)).getpos() );
        Iterator iter=temporaryRemoved.iterator();
	while(iter.hasNext())
	{
	    Koord3D k=(Koord3D)iter.next();
	    this.removeCell(k);
	    markviewd.honeycombHasBeenRemoved(k);
	}
//nderung bez. lschen von markierten Waben
    }
    //stellt alle Waben wieder her die mit selectedRemove gelscht wurden
    public void allTempRemovedBack()
    {
        if(temporaryRemoved.isEmpty())return;
	Iterator iter=temporaryRemoved.iterator();
	while(iter.hasNext())
	     this.addCell((Koord3D)iter.next());
	temporaryRemoved.clear();
    }
    //Alle Waben werden wieder solid dargestellt
    public void allSolid()
    {
        for(int i=0;i<HoneycombAddPoint.numChildren();i++)
        {
            ((Honeycomb) HoneycombAddPoint.getChild(i)).toLines(false);
        }
    }

    public boolean selectPolygon2D(Koord3D[] chosen)
    {
	deselectEverything();
	for(int i=0;i<3;i++)
	 if(chosen[i]==null) return false;
	Vector3d a=new Vector3d(chosen[0].getX(),chosen[0].getY(),chosen[0].getZ());
	Vector3d b=new Vector3d(chosen[1].getX(),chosen[1].getY(),chosen[1].getZ());
	Vector3d c=new Vector3d(chosen[2].getX(),chosen[2].getY(),chosen[2].getZ());
	DistanceMeter measure1=new DistanceMeter(a,b,c);
	Vector3d n=new Vector3d();
	Vector3d ab=new Vector3d(b);
	ab.sub(a);
	Vector3d bc=new Vector3d(c);
	bc.sub(b);
	n.cross(ab,bc);
	Vector3d an=new Vector3d(a);
	an.add(n);
	Vector3d bn=new Vector3d(b);
	bn.add(n);
	DistanceMeter measure2=new DistanceMeter(a,b,bn);
	DistanceMeter measure3=new DistanceMeter(c,a,an);
	DistanceMeter measure4=new DistanceMeter(c,b,bn);
	Honeycomb h;
	Koord3D pos;
	double sgn2=measure2.getDistance(c);
	double sgn3=measure3.getDistance(b);
	double sgn4=measure4.getDistance(a);
	double d2,d3,d4;
	Vector3d vec=new Vector3d();
	for(int i=0;i<HoneycombAddPoint.numChildren();i++)
	{
	       h=(Honeycomb)HoneycombAddPoint.getChild(i);
	       pos=h.getpos();
	       vec.set(pos.getX(),pos.getY(),pos.getZ());
               if(Math.abs(measure1.getDistance(vec))<0.5)
	       {
		    d2=measure2.getDistance(vec);
		    d3=measure3.getDistance(vec);
		    d4=measure4.getDistance(vec);
		    if((sgn2*d2>=0)&&(sgn3*d3>=0)&&(sgn4*d4>=0))
			h.setSelected(true);
	       }
	}
	return true;
    }
    public boolean selectArea(Koord3D[] chosen)
    {
	deselectEverything();
	for(int i=0;i<3;i++)
	 if(chosen[i]==null) return false;
	Vector3d a=new Vector3d(chosen[0].getX(),chosen[0].getY(),chosen[0].getZ());
	Vector3d b=new Vector3d(chosen[1].getX(),chosen[1].getY(),chosen[1].getZ());
	Vector3d c=new Vector3d(chosen[2].getX(),chosen[2].getY(),chosen[2].getZ());
	DistanceMeter measure1=new DistanceMeter(a,b,c);
	Honeycomb h;
	Koord3D pos;
	Vector3d vec=new Vector3d();
	for(int i=0;i<HoneycombAddPoint.numChildren();i++)
	{
	       h=(Honeycomb)HoneycombAddPoint.getChild(i);
	       pos=h.getpos();
	       vec.set(pos.getX(),pos.getY(),pos.getZ());
               if(Math.abs(measure1.getDistance(vec))<0.5)
		    h.setSelected(true);
	}
	return true;
    }
    public boolean selectCube(Koord3D[] chosen)
    {
	deselectEverything();
	for(int i=0;i<2;i++)
	    if(chosen[i]==null) return false;
	Vector3d a=new Vector3d(chosen[0].getX(),chosen[0].getY(),chosen[0].getZ());
	Vector3d b=new Vector3d(chosen[1].getX(),chosen[1].getY(),chosen[1].getZ());
	if( (a.x==b.x) || (a.y==b.y) || (a.z==b.z) )
	    return false;//austieg wegen deformiertem Kubus
	Vector3d c=new Vector3d();
	c.set(a.x,a.y,b.z);
	Vector3d d=new Vector3d();
	d.set(b.x,b.y,a.z);
	Vector3d e=new Vector3d();
	e.set(a.x,b.y,a.z);
	Vector3d f=new Vector3d();
	f.set(b.x,a.y,b.z);
	Vector3d g=new Vector3d();
	g.set(b.x,a.y,a.z);
	Vector3d h=new Vector3d();
	h.set(a.x,b.y,b.z);
	DistanceMeter measure1=new DistanceMeter(a,c,e);
	DistanceMeter measure2=new DistanceMeter(b,h,f);
	DistanceMeter measure3=new DistanceMeter(a,g,c);
	DistanceMeter measure4=new DistanceMeter(b,d,e);
	DistanceMeter measure5=new DistanceMeter(a,g,d);
        DistanceMeter measure6=new DistanceMeter(b,d,f);
	double sgn1=measure1.getDistance(b);
        double sgn2=measure2.getDistance(a);
        double sgn3=measure3.getDistance(b);
        double sgn4=measure4.getDistance(a);
        double sgn5=measure5.getDistance(b);
        double sgn6=measure6.getDistance(a);
	double d1,d2,d3,d4,d5,d6;
	Honeycomb ho;
	Koord3D pos;
	Vector3d vec=new Vector3d();
	for(int i=0;i<HoneycombAddPoint.numChildren();i++)
	{
	    ho=(Honeycomb)HoneycombAddPoint.getChild(i);
	    pos=ho.getpos();
	    vec.set(pos.getX(),pos.getY(),pos.getZ());
	    d1=measure1.getDistance(vec);
	    d2=measure2.getDistance(vec);
	    d3=measure3.getDistance(vec);
	    d4=measure4.getDistance(vec);
	    d5=measure5.getDistance(vec);
	    d6=measure6.getDistance(vec);
	    if((d1*sgn1>=0)&&(d2*sgn2>=0)&&(d3*sgn3>=0)&&(d4*sgn4>=0)&&(d5*sgn5>=0)&&(d6*sgn6>=0))
	     ho.setSelected(true);
	}
	return true;
    }
    public boolean selectPolygon3D(Koord3D[] chosen)
    {
	deselectEverything();
	for(int i=0;i<4;i++)
	 if(chosen[i]==null) return false;
	Vector3d a=new Vector3d(chosen[0].getX(),chosen[0].getY(),chosen[0].getZ());
	Vector3d b=new Vector3d(chosen[1].getX(),chosen[1].getY(),chosen[1].getZ());
	Vector3d c=new Vector3d(chosen[2].getX(),chosen[2].getY(),chosen[2].getZ());
	Vector3d d=new Vector3d(chosen[3].getX(),chosen[3].getY(),chosen[3].getZ());
	DistanceMeter measure1=new DistanceMeter(a,b,c);
	DistanceMeter measure2=new DistanceMeter(a,b,d);
	DistanceMeter measure3=new DistanceMeter(a,d,c);
	DistanceMeter measure4=new DistanceMeter(c,d,b);
	double sgn1=measure1.getDistance(d);
	if(sgn1==0)
	    return false;//Ausstieg weil kein echtes Tetraeder
	double sgn2=measure2.getDistance(c);
	double sgn3=measure3.getDistance(b);
	double sgn4=measure4.getDistance(a);
	double d1,d2,d3,d4;
	Honeycomb h;
	Koord3D pos;
	Vector3d vec=new Vector3d();
	for(int i=0;i<HoneycombAddPoint.numChildren();i++)
	{
	    h=(Honeycomb) HoneycombAddPoint.getChild(i);
	    pos=h.getpos();
	    vec.set(pos.getX(),pos.getY(),pos.getZ());
	    d1=measure1.getDistance(vec);
	    d2=measure2.getDistance(vec);
	    d3=measure3.getDistance(vec);
	    d4=measure4.getDistance(vec);
	    if((sgn1*d1>=0)&&(sgn2*d2>=0)&&(sgn3*d3>=0)&&(sgn4*d4>=0))
		h.setSelected(true);
	}
        return true;
    }

    private void deselectEverything()
    {
	for(int i=0;i<HoneycombAddPoint.numChildren();i++)
		((Honeycomb)HoneycombAddPoint.getChild(i)).setSelected(false);
    }
}
