import java.util.Properties;
import java.io.FileInputStream;
import java.io.FileOutputStream;

/**
 * Title: Parameter
 * Description: Klasse fuer die Speicherung der Dateien.
 * @author Daniel Schlosser & Dimitri Missoh
 * @version 0.1
 */

public abstract class Parameters {

    private boolean DEBUG = false;
    private String propertiesFilename;
    private String propertiesDescription;

    protected Properties properties = null;

    protected Parameters(String propertiesFilename, String propertiesDescription) {
        this.propertiesFilename = propertiesFilename;
        this.propertiesDescription = propertiesDescription;
    }

    abstract protected void setDefaults(Properties defaults) ;
    abstract protected void updatePropertiesFromSettings() ;
    abstract protected void updateSettingsFromProperties() ;

    protected void getParameters() {
        Properties defaults = new Properties();
        FileInputStream in = null;

        setDefaults(defaults);

        properties = new Properties(defaults);

        try {
            String folder = System.getProperty("user.home");
            String filesep = System.getProperty("file.separator");
            in = new FileInputStream(folder
                                     + filesep
                                     + propertiesFilename);
            properties.load(in);

        } catch (java.io.FileNotFoundException e) {
            in = null;
            ErrorMessages.error("Einstellungsdatei nicht gefunden. " +
                                "Die vordefinierten Werte werden benutzt.");
        } catch (java.io.IOException e) {
            ErrorMessages.error("Einstellungsdatei kann nicht gelesen werden. " +
                                "Die vordefinierten Werte werden benutzt.");
        } finally {
            if (in != null) {
                try { in.close(); } catch (java.io.IOException e) { }
                in = null;
            }
        }

        updateSettingsFromProperties();

    }//Ende getParameter()

    protected void saveParameters() {

        updatePropertiesFromSettings();

        if (DEBUG) {
            System.out.println("Die Einstellungen werden eingesetzt: " + propertiesDescription);
            System.out.println(toString());
        }

        FileOutputStream out = null;

        try {
            String folder = System.getProperty("user.home");
            String filesep = System.getProperty("file.separator");
            out = new FileOutputStream(folder
                                       + filesep
                                       + propertiesFilename);
            properties.store(out, propertiesDescription);
        } catch (java.io.IOException e) {
            ErrorMessages.error("Kann nicht die Einstellungen speichern");
        } finally {
            if (out != null) {
                try { out.close(); } catch (java.io.IOException e) { }
                out = null;
            }
        }
    }//Ende saveParameter()
}

