/** Wir verwenden teilweise nicht die Random-Klasse aus der Sun-API. 
 * Anstelle dessen verwenden wir diese Klasse, geklaut von <a href="http://www.theorem.com/java/Mersenne.java">Mersenne Twister</a>.
 * Sie ist ein bisschen schneller und liefert mindestens genauso zufllige Zahlen. Allerdings ist ihr Handling ein bisschen komplizierter.<br><br>
 * Methoden werden nicht kommentiert...
 *
 * Stand: 09.07.2001
 */

public class Random
{
        // Period parameters
	private static final int N = 624;
	private static final int M = 397;
	private static final int MATRIX_A = 0x9908b0df;   //	private static final * constant vector a
	private static final int UPPER_MASK = 0x80000000; // most significant w-r bits
	private static final int LOWER_MASK = 0x7fffffff; // least significant r bits


	// Tempering parameters
	private static final int TEMPERING_MASK_B = 0x9d2c5680;
	private static final int TEMPERING_MASK_C = 0xefc60000;
	private static final int UMASK = 0xffffffff;	// unsigned mask for promoting int -> long

	// #define TEMPERING_SHIFT_U(y)  (y >>> 11)
	// #define TEMPERING_SHIFT_S(y)  (y << 7)
	// #define TEMPERING_SHIFT_T(y)  (y << 15)
	// #define TEMPERING_SHIFT_L(y)  (y >>> 18)

	private int mt[]; // the array for the state vector
	private int mti; // mti==N+1 means mt[N] is not initialized
	private int mag01[];


	/**
	* Constructor using the default seed.
	*/
	public Random()
	{
		sgenrand(4357);
	}

	/**
	* Constructor using a given seed.
	*
	* @param seed generator starting number, often the time of day.
	*/
	public Random(int seed)
	{
		sgenrand(seed);
	}

	/**
	* Initalize the pseudo random number generator.
	*
	* @param seed from constructor
	*
	*/
	private void sgenrand(int seed)
	{
		mt = new int[N];

		// setting initial seeds to mt[N] using
		// the generator Line 25 of Table 1 in
		// [KNUTH 1981, The Art of Computer Programming
		//    Vol. 2 (2nd Ed.), pp102]
		mt[0]= seed & 0xffffffff;

		for (mti = 1; mti < N; mti++)
			mt[mti] = (69069 * mt[mti-1]) & 0xffffffff;

		// mag01[x] = x * MATRIX_A  for x=0,1
		mag01 = new int[2];
		mag01[0] = 0x0;
		mag01[1] = MATRIX_A;
	}

	/**
	* Generate the next random number.
	*/
	public int genrand()
	{
		int y;

		if (mti >= N)	// generate N words at one time
		{
			int kk;

			for (kk = 0; kk < N - M; kk++)
			{
				y = (mt[kk] & UPPER_MASK) | (mt[kk+1] & LOWER_MASK);
				mt[kk] = mt[kk+M] ^ (y >>> 1) ^ mag01[y & 0x1];
			}
			for (; kk < N-1; kk++)
			{
				y = (mt[kk] & UPPER_MASK) | (mt[kk+1] & LOWER_MASK);
				mt[kk] = mt[kk+(M-N)] ^ (y >>> 1) ^ mag01[y & 0x1];
			}
			y = (mt[N-1] & UPPER_MASK) | (mt[0] & LOWER_MASK);
			mt[N-1] = mt[M-1] ^ (y >>> 1) ^ mag01[y & 0x1];

			mti = 0;
		}

		y = mt[mti++];
		y ^= y >>> 11;								// TEMPERING_SHIFT_U(y)
		y ^= (y << 7) & TEMPERING_MASK_B;	// TEMPERING_SHIFT_S(y)
		y ^= (y << 15) & TEMPERING_MASK_C;	// TEMPERING_SHIFT_T(y)
		y ^= (y >>> 18);							// TEMPERING_SHIFT_L(y)

		return y;
	}
}
